/////////////////////////////////////////////////////////////////////
//Note, this code was originally written by Matt Weir, at Florida State University
//Contact Info weir -at- cs.fsu.edu
//This code is licensed GPLv2
//If you make any changes I would love to hear about them so I don't have to impliment them myself ;)
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <math.h>


//Note, the word DIC is used quite often throughout this code to represent dictionary
//I know, it also sounds like a part of the male anatomy, hahaha, get your laughter out now
//I used that name just because it's shorter than word-list
//BTW if you can make any inappropriate jokes involving "word-list" more power to you
//Aka, Is it just cold or do you have a really small wordlist?
//Hey, I didn't say they had to be good, or funny, give me a break here

#define MAXSECTIONS 16  //Maximum number of sections for a grammar
#define MAXWORDSIZE 16 //Maximum size of a generated password
#define MAXDICCAT 10      //Maximum number of input dictionary categories, Note if more than 10, may need to do some recoding
#define MAXINPUTDIC 10  //Maximum number of user inputed dictionaries
#define MAXDICNAME 40   //Maximum length of the input dictionary name

////////////////////////////////////////////////
//The individual sections
typedef struct secStruct {
  char word[MAXWORDSIZE+1];  //the section info, aka for a 4 digit it might equal 1234
  int Cat;                   //used for dic words, to quickly figure out what catagory it is
  double prob;                  //the probability of this section appearing
  secStruct* next;           //the next hightest prob section
}secType;


///////////////////////////////////////////////
//Node in the priority quene
typedef struct orderStruct {
  int curPivot;                    //What value it can pivot on
  int size;                        //The number of sections in this structure
  char secDef[MAXSECTIONS];        //What each section is, aka L=lowercase, U=uppercase, S=special char, D=digit
  int secSize[MAXSECTIONS];       //How big each section is, aka 4 digits, 2 special chars, etc.
  int dicPointer[MAXSECTIONS];     //used to keep track of which dictionary it is on
  secType *secPointer[MAXSECTIONS];//Points to the current sections
  float prob;                      //The probability of the terminal structure
  float strucProb;		   //The probability of the base structure
  orderStruct *next;
  orderStruct *prev;
}orderType;

////////////////////////////////////////////
//Holds the dictionary words
typedef struct dicHolderStruct {
  char word[MAXWORDSIZE+1];
  dicHolderStruct *next;
}dicHolderType;



secType *number_sec[MAXWORDSIZE+1];    //holds all the number probabilities
secType *special_sec[MAXWORDSIZE+1];   //holds all the special character probabilities
secType *word_sec[MAXWORDSIZE+1];
orderType *pQueue;
int totalWords[MAXDICCAT][MAXWORDSIZE+1];
dicHolderType *dicWords[MAXDICCAT][MAXWORDSIZE+1];

//--For Debugging purposes these values are used, (sounds like Yoda I do)
int curCountQueue;
int maxCountQueue;
int totalRounds;
//--End of debugging values

void sig_alrm(int signo)
{
  printf("exiting\n");
  exit(0);
}

int initilize();				//Main fuction used to initilize variables
int initilize_secStruct(FILE *inputFile,int i,int type);	//Initilizes all the section structures
int initilize_secWord(char tempCatInfo,float *tempWordProb, int tempCat);    //Initilizes the word structures;
int initilize_pQueue(FILE *inputFile);	//Initilizes the priority queue with the intitial base structures
orderType* initilize_orderType();				//Initilizes a new orderType
orderType* setStruct_orderType(orderType *newPointer,char *tempWord);
int insertQueue(orderType *newPointer);    //Inserts a node into the priority queue
int pop(int outputType);		//pops the top value out of the priority queue, inserts new values if generated, returns the terminal structure of the popped
						//value
int printOrder(orderStruct *dataValue, int outputType);		//prints the final terminal structure info to stdout
int addChildren(orderStruct *dataValue);	//Adds children of the node to the priority queue
orderType* copyOrderType(orderType *dataValue); //makes a direct copy of an orderType structure, (except for the prev/next nodes which are left NULL);
double calcProb(orderType *newPointer);		//Calculates the probability of the structure
char letters[27] ="abcdefghijklmnopqrstuvwxyz";  //Used to denote all the possible alpha letters that are possible, used to support other languages in the future


//----------The following functions are used for processing the dictionary files and inserting the words
//----------This distinction is made to make life easier to multi-thread/ or distribute this to other computers in the future
int readDic(char *userDic,int userDicLevel);
int getNextWord(FILE *inputFile,char *nextWord);
int convertLower(char *inputPass);
int fillInWord(orderType* dataValue, char *printValue, int startPos);


int main(int argc, char *argv[]) {
  struct sigaction action;
  int badCommandLine;    //True if improper values are passed via the command line
  int outputType;        //What to output, finished password guesses, terminal structures, or size of priority queue
  char userDic[MAXINPUTDIC][MAXDICNAME];  //the user supplied input dictionaries
  int userDicLevel[MAXINPUTDIC];          //the "value" of the user dictionary  
  int i;
  int j;
  int numUserDic;       //The number of dictionaries the user entered
  secStruct *tempSec;
//---------Parse the command line------------------------//
  badCommandLine =0;
  outputType = 0;
  numUserDic=0;
  if (argc == 1) {
    badCommandLine=1;
  }
  for (i=1;i<argc;i++) {
    if (argv[i][0]!='-') {
      badCommandLine=1;
      printf("%s",argv[i]);
      break;
    }
    else if ((argv[i][1]!='d')&&(argv[i][1]!='s')&&(argv[i][1]!='q')) {
      badCommandLine=1;
      break;
    }
    else if (argv[i][1]=='q') {
      if (outputType!=0) {  //if a different flag was already specified
        badCommandLine=1;
        printf("2");
        break;
      }
      else {
        outputType=2;
      }
    }
    else if (argv[i][1]=='s') {
      if (outputType!=0) {  //if a different flag was already specified
        badCommandLine=1;
        printf("3");
        break;
      }
      else {
        outputType=1;
      }
    }
    else if (argv[i][1]=='d') {
      if (i==argc-1) { //there isn't a dictionary specified and/or the level isn't specified
        badCommandLine=1;
        printf("4");
        break;
      }
      else if (!isdigit(argv[i][2])) {
        badCommandLine=1;
        printf("5");
        break;
      }      
      else {
        if (numUserDic<MAXINPUTDIC) {
          userDicLevel[numUserDic]=atoi(argv[i]+2);
          strncpy(userDic[numUserDic],argv[i+1],MAXDICNAME);
          numUserDic++;
          i++;  //Don't parse the next input since it is the dictionary name
        }
        else {
          printf("Sorry you inputed too many dictonaries.  The maximun number supported is %i\n",MAXINPUTDIC);
          exit(0);
        } 
      }
    } 
  }
  if (numUserDic<1) {   //No dictionaries were entered
    badCommandLine=1;
    printf("6");
  }
  if (badCommandLine) {
    printf("Usage: ./pcfg_manager -d[0-9] <dictionaryfile>\n");
    printf("Aka    ./pcfg_manager -d0 gooddictionary.txt -d8 poordictionary.txt\n");
    printf("Or     ./pcfg_manager -s -d0 gooddictionary.txt   to print only the terminal structures\n");
    printf("Or     ./pcfg_manager -q -d0 gooddictionary.txt  to print the size of the priority queue\n");
    exit(0);
  }

//-----------------End Parsing the Command Line------------//
//-----------------Signal Initilization--------------
  action.sa_handler = sig_alrm;
  sigemptyset(&action.sa_mask);
  action.sa_flags = 0;
  sigaction(SIGINT, &action, 0);
  sigaction(SIGKILL, &action, 0);
  sigaction(SIGTERM, &action, 0);
  sigaction(SIGALRM, &action, 0);

//-----------------End Signal Initilization----------
  curCountQueue=0;
  maxCountQueue=0;
  totalRounds++;
  for (i=0;i<=MAXWORDSIZE;i++) {
    for (j=0;j<MAXDICCAT;j++) {
      totalWords[j][i]=0;
    }
  }
  for (i=0;i<numUserDic;i++) {
    if (readDic(userDic[i], userDicLevel[i])!=0) {
      printf("Could not open the dictionary file %s\n",userDic[i]);
      return -1;
    }
  }
//-----used for debugging
//  for (i=0;i<3;i++) {
//    for (j=0;j<=MAXWORDSIZE;j++) {
//      printf("Number Cat %i Size %i = %i\n",i,j,totalWords[i][j]);
//    }    
//  }
  if (initilize()==-1) {
    return -1;
  }
//  for (i=1;i<=MAXWORDSIZE;i++) {
//    tempSec=word_sec[i];
//    while (tempSec!=NULL) {
//      printf("size=%i cat=%s\n",i, tempSec->word);
//      tempSec=tempSec->next;
//    }
//    printf("size=%i NULL\n",i); 
//  }
  printf("starting to generater passwords\n");
  while (pQueue!=NULL) {
    pop(outputType);
  }  
}

///////////////////////////////////////////////////////////////
//Initilizes all the variables
//
int initilize () {
  FILE *inputFile;
  char baseDir[100];
  char fileName[150];
  int i;
  int j;
  int tempMult[MAXDICCAT];
  char tempCatInfo[MAXDICCAT+1]="abcdefghij";  //used to tell which word category to use, yes it seems stupid but it really helps with debugging
  float tempWordProb[MAXDICCAT][MAXWORDSIZE+1];
  pQueue=NULL;
  for (i=0;i<MAXWORDSIZE;i++) {
    number_sec[i]=NULL;
    special_sec[i]=NULL;
  }

  //-------Populate the digits--------------//
  strncpy(baseDir,"./digits/",99);
  for (i=1;i<=MAXWORDSIZE;i++) {
    snprintf(fileName,149,"%s%i.txt",baseDir,i);
    inputFile=fopen(fileName,"r");
    if (inputFile!=NULL) {
      if (initilize_secStruct(inputFile,i,0)==-1) {
        return -1;
      }
      fclose(inputFile);
    }
  }

  //-------Populate the special characters---//
  strncpy(baseDir,"./special/",99);
  for (i=1;i<=MAXWORDSIZE;i++) {
    snprintf(fileName,149,"%s%i.txt",baseDir,i);
    inputFile=fopen(fileName,"r");
    if (inputFile!=NULL) {
      if (initilize_secStruct(inputFile,i,1)==-1) {
        return -1;
      }
      fclose(inputFile);
    }
  } 

  //--------Initilize the word probabilities--//
  //--------Note, doesn't give TRUE probablities with more than one dictionary, change that in next version--//
  for (i=0;i<MAXDICCAT;i++) {  //set up the multiplication factors for each dic group.  Each group is 1/2 as likely as the one before
    tempMult[i]=pow(2,i);      //As I said, not a true probability if you have more than one cat, but it should be good enough.
  }
  for (i=0;i<=MAXWORDSIZE;i++) {
    word_sec[i]=NULL;
  }
  for (i=0;i<MAXDICCAT;i++) {  //find the initial probabilities
    for (j=1;j<=MAXWORDSIZE;j++) {
      if (totalWords[i][j]!=0) {
        tempWordProb[i][j]=1.0/totalWords[i][j] * (1.0/tempMult[i]);
      }
      else {
        tempWordProb[i][j]=0;
      }
    }
  }
  for (i=0;i<MAXDICCAT;i++) {
    initilize_secWord(tempCatInfo[i],tempWordProb[i], i);
  }
  //--------Now initilize the structures and insert them in the priority queue-----//
  inputFile=fopen("./grammar/structures.txt","r");
  if (inputFile==NULL) {
    return -1;
  }
  pQueue=NULL;
  if (initilize_pQueue(inputFile)==-1) {
    return -1;
  }
  return 0;
}





//////////////////////////////////////////////////////////////
//Initilizes the secstructure that holds number and special character probabilities
int initilize_secStruct(FILE *inputFile,int i,int type) {
  char tempChar;
  char tempWord[MAXWORDSIZE+30];
  secType *secPointer;
  int j;

  if (type ==0) { //is a number
    secPointer=number_sec[i];
  } 
  else if (type==1) {
    secPointer=special_sec[i];
  }
  else { 
    return -1;
  }
  tempChar=fgetc(inputFile);
  while (!feof(inputFile)) {
    //-------First get the digits---------//
    j=0;
    while ((!feof(inputFile))&&(tempChar!='\t')) {
      tempWord[j]=tempChar;
      tempChar=fgetc(inputFile);
      if (j<=i) {
        j++;
      }
      else {
        printf("error in processing digit file\n");
        return -1;
      }
    }
    tempWord[j]='\0';
    if (secPointer!=NULL) {
      secPointer->next=new secType;
      secPointer=secPointer->next;
    }
    else {
      secPointer=new secType;
      if (type ==0) { //a number
        number_sec[i]=secPointer;
      }
      else if (type ==1) { //a special char
        special_sec[i]=secPointer;
      }
    }
    strncpy(secPointer->word,tempWord,MAXWORDSIZE+1);

    j=0;
    //----Now get the percentage-----//
     tempChar=fgetc(inputFile);
     while ((!feof(inputFile))&&(tempChar!='\n')) {
       tempWord[j]=tempChar;
       tempChar=fgetc(inputFile);
       if (j<MAXWORDSIZE+30) {
         j++;
       }
       else {
         printf("error in processing percentages in digit file\n");
         return -1;
       }
    }
    tempWord[j]='\0';
    secPointer->prob=atof(tempWord);
    secPointer->next=NULL;
    j=0;
    tempChar=fgetc(inputFile);
  }
  return 0;
}


///////////////////////////////////////////////////////////////
//New
//Initilizes the word structures, to make them similar to all the other structures
//We still treat them a little differently to keep the size of the priority queue sane
int initilize_secWord(char tempCatInfo,float *tempWordProb, int tempCat) {
  char tempChar;
  char tempWord[MAXWORDSIZE+30];
  secType *secPointer;
  secType *tempPointer;
  int i;
  int j;
  
  for (i=0;i<=MAXWORDSIZE;i++) {
    if (tempWordProb[i]!=0) {  //only insert nodes if they have actual words in the dictionary for that length
      secPointer=word_sec[i];
      if (secPointer!=NULL) {  //Handle inserting the new node/where to insert it
        while ((secPointer->next!=NULL)&&(secPointer->next->prob>tempWordProb[i])) {
          secPointer=secPointer->next;
        }
        tempPointer = new secType;
        if (secPointer->next!=NULL) {
          tempPointer->next=secPointer->next;
          secPointer->next=tempPointer;
        }
        else {
          if (secPointer->prob >= tempWordProb[i]) {
            secPointer->next=tempPointer;
            tempPointer->next=NULL;
          }
          else {
            tempPointer->next=secPointer;
            word_sec[i]=tempPointer;
          }
        }
        secPointer=tempPointer;
      }
      else {
        secPointer=new secType;
        secPointer->next=NULL;
        word_sec[i]=secPointer;
      }
      for (j=0;j<i;j++) {    //insert the full marker, (once again, this may look stupid, but it really helps in debugging with the -s option
        secPointer->word[j]=tempCatInfo;
      }
      secPointer->Cat=tempCat; //added to support multiple dictionaries
      secPointer->word[i]='\0';
      
      secPointer->prob=tempWordProb[i];
    }
  }
  return 0;
} 





int initilize_pQueue(FILE *inputFile){
  char tempChar;
  char tempWord[MAXWORDSIZE+30];
  orderType *newPointer;
  int i;
  int j;
  int badInput;

  tempChar=fgetc(inputFile);
  while (!feof(inputFile)) {
    //-------First get the structures---------//
    i=0;
    badInput=0;
    while ((!feof(inputFile))&&(tempChar!='\t')) {
      tempWord[i]=tempChar;
      tempChar=fgetc(inputFile);
      if (i<=MAXWORDSIZE) {
        i++;
      }
      else {
        badInput=1;
      }
    }
    tempWord[i]='\0';
    if (badInput!=1) {
      newPointer= initilize_orderType();
      newPointer= setStruct_orderType(newPointer,tempWord);
      if (newPointer==NULL) {
        badInput=1;
      }
    }
    i=0;
    //----Now get the percentage-----//
    tempChar=fgetc(inputFile);
    while ((!feof(inputFile))&&(tempChar!='\n')) {
      tempWord[i]=tempChar;
      tempChar=fgetc(inputFile);
      if (i<MAXWORDSIZE+30) {
        i++;
      }
      else {
        printf("error in processing percentages in structure file\n");
        return -1;
      }
    }
    tempWord[i]='\0';
    if (badInput!=1) {
      newPointer->strucProb=atof(tempWord);
      //------Now calculate the total probability---------//
      newPointer->prob=calcProb(newPointer);
      //-----Now insert into priority queue---------//
      if (newPointer->prob!=0) {
        insertQueue(newPointer);
      }
    }
    i=0;
    tempChar=fgetc(inputFile);
  }
  return 0;
}

orderType * initilize_orderType() {
  orderType *newPointer;
  int i;
  
  newPointer= new orderType;
  newPointer->next=NULL;
  newPointer->prev=NULL;
  newPointer->curPivot=0;
  newPointer->size=0;
  newPointer->prob=0;
  newPointer->strucProb=0;
  for (i=0;i<MAXSECTIONS;i++) {
    newPointer->dicPointer[i]=0;
    newPointer->secDef[i]=' ';
    newPointer->secSize[i]=0;
    newPointer->secPointer[i]=NULL;
  }
  return newPointer;
}


orderType *setStruct_orderType(orderType* newPointer,char* tempWord) {
  int size;
  int j;
  int m,n;
  int pos;
  char tempChar;
  
  j=0;
  pos=0;
  size=1;
  tempChar=tempWord[pos];
  while (tempChar!='\0') {
    pos++;
    if (tempChar==tempWord[pos]){
      size++;
    }
    else {
      newPointer->secDef[j]=tempChar;
      newPointer->secSize[j]=size;
      if (tempChar=='S') {
        newPointer->secPointer[j]=special_sec[size];
      }
      else if (tempChar=='D') {
        newPointer->secPointer[j]=number_sec[size];
      }
      else if (tempChar=='L'){  //if a word
        newPointer->secPointer[j]=word_sec[size];
      }
      else {  //if invalid
        printf("ERROR\n");
        return NULL;
      }
      j++;
      if (j>=MAXSECTIONS) {
        return NULL;
      }
      tempChar=tempWord[pos];
      size=1;
    }   
  }
  newPointer->size=j;

  return newPointer;
}

int insertQueue(orderType *newPointer){
  orderStruct *tempPointer;

  curCountQueue++;
  if (curCountQueue>maxCountQueue) maxCountQueue=curCountQueue;
  tempPointer=pQueue; 
  if (pQueue==NULL) {   //if the priority queue is empty
    pQueue=newPointer;
    newPointer->prev=NULL;
    newPointer->next=NULL;
  }
  else if (newPointer->prob>=tempPointer->prob) {  //if the new value is the first value
    newPointer->prev=NULL;
    newPointer->next=pQueue;
    pQueue->prev=newPointer;
    pQueue=newPointer;
  }
  else {  //otherwise search the queue to insert the new value
    while ((tempPointer->next!=NULL)&&(tempPointer->next->prob>newPointer->prob)) {
      tempPointer=tempPointer->next;
    }
    if (tempPointer->next==NULL) {
      newPointer->next=NULL;
      newPointer->prev=tempPointer;
    }
    else {
      newPointer->next=tempPointer->next;
      tempPointer->next->prev=newPointer;
    }
    tempPointer->next=newPointer;
  }
  return 0;
}

////////////////////////////////////////////////////////
//Pops the top value out of the priority queue
//Prints the top value
//Inserts any child nodes back into the priority queue
//
//outputType
//0=output the full password guess
//---the other options are for debugging and research purposes
//1=output the terminal structure, (before the dictionary words are inserted
//2=output the size of the priority queue instead
int pop(int outputType) {
  orderType *tempPointer;
  if (pQueue==NULL) {
    printf("Error, pQueue is NULL\n");
    return -1;
  }
  totalRounds++;
  curCountQueue--;
  printOrder(pQueue,outputType);
  tempPointer=pQueue;
   
  //--Delete the node--//
  if (pQueue->next!=NULL) { 
    pQueue=pQueue->next;
    pQueue->prev=NULL;
  }
  else {
    pQueue=NULL;
  }
  //--Add Children--//
  addChildren(tempPointer);
  delete tempPointer;

  return 0;
}


int printOrder(orderStruct *dataValue, int outputType) {
  int i;
  int j;
  int rounds;
  char printValue[MAXWORDSIZE+4];

  if (dataValue==NULL) {
    printf("Error dataValue is NULL\n");
    return -1;
  }
 
  printValue[0]='\0';
  if (outputType==1) {
    for (i=0;i<dataValue->size;i++) {
      strncat(printValue,dataValue->secPointer[i]->word,MAXWORDSIZE);
    }
    printf("%s %1.10f\n",printValue,dataValue->prob); 
  }
  else if (outputType==2) {   
    printf("Current=%i Max=%i TotalRounds=%i\n",curCountQueue,maxCountQueue,totalRounds);      
  }
  else if (outputType==0) {  //---fill in the dictionary words
    fillInWord(dataValue,printValue,0);
  } 
  return 0;
}

int addChildren(orderType *dataValue) {
  orderType *tempPointer; 
  int i;

  for (i=dataValue->curPivot;i<dataValue->size;i++) {
    if (dataValue->secPointer[i]->next!=NULL) {  //Can pivot on that point
      tempPointer=copyOrderType(dataValue);
      tempPointer->secPointer[i]=tempPointer->secPointer[i]->next;
      tempPointer->curPivot=i;
      tempPointer->prob=calcProb(tempPointer);
      if (tempPointer->prob!=0) {
        insertQueue(tempPointer); 
      }
    }
  }
  return 0;
} 
 

orderType* copyOrderType(orderType *dataValue) {
  orderType *tempPointer;
  int i;

  tempPointer=initilize_orderType();
  tempPointer->curPivot= dataValue->curPivot;
  tempPointer->size=dataValue->size;
  tempPointer->prob=dataValue->prob;
  tempPointer->strucProb=dataValue->strucProb;
  for (i=0;i<dataValue->size;i++) {
    tempPointer->dicPointer[i]=dataValue->dicPointer[i];
    tempPointer->secDef[i]=dataValue->secDef[i];
    tempPointer->secSize[i]=dataValue->secSize[i];
    tempPointer->secPointer[i]=dataValue->secPointer[i];
  }

  return tempPointer;
}

double calcProb(orderType *newPointer) {
  double finalValue; 
  int j; 
  finalValue=newPointer->strucProb;
  for(j=0;j<newPointer->size;j++) {
    if (newPointer->secPointer[j]==NULL) {
      return 0;
    }
    finalValue=finalValue*newPointer->secPointer[j]->prob;
  }
  return finalValue;
}

int readDic(char *userDic,int userDicLevel) {
  FILE *inputFile;
  char nextWord[MAXWORDSIZE+1];
  int done;
  char curChar;
  int i;
  dicHolderType *curPos;
  dicHolderType *tempWord;
  int size;
  int isNew;    //used to tell if the word is new 
  inputFile=fopen(userDic,"r");
  if (inputFile==NULL) {
    return -1;
  }  
  done=getNextWord(inputFile,nextWord);
  size = strlen(nextWord);
  while (done!=1) {
    isNew=0;  //assume the word is not new
    if (done!=-1) {  //if the dictionary word is valid
      curPos=dicWords[userDicLevel][size];
      //-----------------------------Note, does not check for duplicates yet, add in the next version---------///
      if (curPos==NULL) {
         curPos=new dicHolderType;
         curPos->next=NULL;
         strncpy(curPos->word,nextWord,MAXWORDSIZE);
         dicWords[userDicLevel][size]=curPos;
      }
      else {
        while (curPos->next!=NULL) {
          curPos=curPos->next;
        }
        tempWord=new dicHolderType;
        curPos->next=tempWord;
	tempWord->next=NULL;
        strncpy(tempWord->word,nextWord,MAXWORDSIZE);
      }
      isNew=1;  //yes, not checking for duplicates just yet, delete this in the next version
 
      if (isNew) {   //increment the word count
        totalWords[userDicLevel][size]++;
      }
    }
    done=getNextWord(inputFile,nextWord);
    size=strlen(nextWord);
  }
  fclose(inputFile);
  return 0;
}


//////////////////////////////////////////////////////////
//gets the next word from the inputfile
//only processes alpha characters
//converts all characters to lower case
//returns 0 if everything works ok
//returns 1 if end of file
//returns -1 if the word is not valid
int getNextWord(FILE *inputFile,char *nextWord) {
  char tempChar;
  int curCount=0;

  tempChar=fgetc(inputFile);
  convertLower(&tempChar);
  while(!feof(inputFile)&&(tempChar!='\n')&&(tempChar!='\r')) {
    if (strchr(letters,tempChar)==NULL) {
      return -1;
    }
    if (curCount < (MAXWORDSIZE)) {
      nextWord[curCount]=tempChar;
      curCount++;
    }
    else {
      return -1;
    }
    tempChar=fgetc(inputFile);
    convertLower(&tempChar);
  }
  nextWord[curCount]='\0';  
  if (feof(inputFile)){
    return 1;
  }
  else {
    return 0;
  }
} 


////////////////////////////////////////////////////////////
//Included this as a seperate function so I can handle other languages in the future
//Currently it is just set up to do English, which is why it's so trivial
int convertLower(char *inputPass) {
  if (isupper(inputPass[0])) {
    inputPass[0]=tolower(inputPass[0]);
    return 1;
  }
  return 0;
}


int fillInWord(orderType* dataValue, char *printValue, int startPos) {
  dicHolderType *curPos;
  char tempString[MAXWORDSIZE+4];

  if (startPos==dataValue->size) { //done processing
    return 0;
  }
  if (dataValue->secDef[startPos]!='L') {     //if not a letter combo
    strncat(printValue,dataValue->secPointer[startPos]->word,MAXWORDSIZE);
    if (startPos==dataValue->size-1) {
      printValue[MAXWORDSIZE]='\0';
      printf("%s\n",printValue);
    }
    else {
      fillInWord(dataValue,printValue,startPos+1);
    }
  }
  else {  //if it is a dictionary word
     curPos=dicWords[dataValue->secPointer[startPos]->Cat][dataValue->secSize[startPos]];
     while (curPos!=NULL) {
       strncpy(tempString,printValue,MAXWORDSIZE);
       strncat(tempString,curPos->word,MAXWORDSIZE);
       if (startPos==dataValue->size-1) {
	 tempString[MAXWORDSIZE]='\0';
         printf("%s\n",tempString);
       }
       else {
         fillInWord(dataValue,tempString,startPos+1);
       }
       curPos=curPos->next;
     }
  }
  return 0;
} 
