/////////////////////////////////////////////////////////////////////
//Note, this code was originally written by Matt Weir, at Florida State University
//Contact Info weir -at- cs.fsu.edu
//This code is licensed GPLv2
//If you make any changes I would love to hear about them so I don't have to impliment them myself ;)

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define Max_Input_Size 8000 //the number of special char combos
#define Max_Words 100000  //the number of passwords analyzed
#define Pass_Length 14    //the max size of a password to analyze
#define Max_Error 500     //the max size to break passwords up
			 //aka if it's really long password


typedef struct dataStruct {
  char startSpecial[Max_Input_Size][Pass_Length+1];
  char middleSpecial[Max_Input_Size][Pass_Length+1];
  char endSpecial[Max_Input_Size][Pass_Length+1];
  char numbers[Max_Words][Pass_Length+1];
  char words[Max_Words][Pass_Length+1];
  char layout[Max_Words][Pass_Length+1];
  int layoutCount[Max_Words];
  int startCount[Max_Input_Size];
  int middleCount[Max_Input_Size];
  int endCount[Max_Input_Size];
  int wordCount[Max_Words];
  int numberCount[Max_Words];
  int numNumbers;
  int numStart;
  int numMiddle;
  int numEnd;
  long numLayout;
  long numWords;
} dataType;

int printOnlyWords;


int parseInput(dataType * passData, char * nextString);
int getNextString(FILE * sysInput, char * nextString);
void initilize (dataType * passData);
void insertString(dataType * passData, char *nextString, int pos, char wType);
void printOutput(FILE * startFile, FILE * middleFile, FILE * endFile, FILE* wordFile,FILE* layoutFile,FILE * numberFile,dataType * passData); 


int main(int argc, char *argv[]) {


  FILE * inputFile;
  FILE * startOutput;
  FILE * middleOutput;
  FILE * endOutput;
  FILE * wordsOutput;
  FILE * rejectOutput;
  FILE * layoutOutput;
  FILE * numberOutput;
  dataType passData;
  int goodBad = 0;
  int totalCount=0;
  char nextString[Max_Error];
  int i = 0;
  if ((argc != 2)&&(argc !=3)) {
    printf("Usage: ./parsePass <inputFile>\n");
    printf("Or     ./parsePass -o <inputFile> to print out a wordlist without number of occurences\n");
    exit(0);
  }
  if (argc==2) {
    inputFile=fopen(argv[1],"r");
    printOnlyWords=0;
    if (inputFile==NULL) {
      printf("ERROR Inputfile NOT FOUND\n");
      return 1;
    }
  }
  else {
    printOnlyWords=1;
    inputFile=fopen(argv[2],"r");
    if (inputFile==NULL) {
      printf("ERROR Inputfile NOT FOUND\n");
      return 1;
    }
  }
  startOutput=fopen("startResults.txt","w");
  if (startOutput==NULL) {
    printf("ERROR can not open startResults.txt for writing\n");
    return 1;
  }
  middleOutput=fopen("middleResults.txt","w");
  if (middleOutput==NULL) {
    printf("ERROR can not open middleResults.txt for writing\n");
    return 1;
  }
  endOutput=fopen("endResults.txt","w");
  if (endOutput==NULL) {
    printf("ERROR can not open endResults.txt for writing\n");
    return 1;
  }
  wordsOutput=fopen("wordsResults.txt","w");
  if (wordsOutput==NULL) {
    printf("ERROR can not open wordsResults.txt for writing\n");
    return 1;
  }
  rejectOutput=fopen("rejectResults.txt","w");
  if (rejectOutput==NULL) {
    printf("ERROR can not open rejectResults.txt for writing\n");
    return 1;
  }
  layoutOutput=fopen("layoutResults.txt","w");
  if (layoutOutput==NULL) {
    printf("ERROR can not open layoutResults.txt for writing\n");
    return 1;
  }
  numberOutput=fopen("numberResults.txt","w");
  if (numberOutput==NULL) {
    printf("ERROR can not open numberResults.txt for writing\n");
    return 1;
  }



  initilize(&passData);
  while (!feof(inputFile)) {
    totalCount++;
    goodBad=getNextString(inputFile, nextString);
    if (goodBad==0) {  /*IF a rejected password*/
      fprintf(rejectOutput,"%s\n",nextString); 
    }
    else if (goodBad==1) { /*If a vailid password*/
      parseInput(&passData,nextString);
    }
  }
  printf("start=%i, middle=%i, end=%i words=%i numLayout=%u total=%u\n",passData.numStart,passData.numMiddle,passData.numEnd,passData.numWords, passData.numLayout,totalCount);

  printOutput(startOutput,middleOutput,endOutput,wordsOutput,layoutOutput,numberOutput,&passData);

  return 0;
} 


/////////////////////////////////////////////////////////////////
//
// Grabs the next password to parse
//
// Returns
//  1 if success
//  0 if it is an invalid password, (contains a space or too long)
// -1 if no more passwords to parse
//
int getNextString(FILE * inputFile, char * nextString){
  char tempChar;
  int curCount=0;
  int badData=0;
 
  tempChar=fgetc(inputFile);
  while(!feof(inputFile)&&(tempChar!='\n')) {
    curCount++;
    if (tempChar==' ') {
      badData=1;
    }
    if (curCount < (Max_Error-1)) {
      nextString[curCount-1]=tempChar;
    }
    tempChar=fgetc(inputFile);
  }
  nextString[curCount]='\0';  
  if ((badData==1)||(curCount > Pass_Length)){
    return 0;
  }
  else if (curCount ==0) {
    return -1;
  }
  else {
    return 1;
  }
}



////////////////////////////////////////////////////////////////
//
// Initilizes all the data structures
//
//
void initilize (dataType * passData) {
  int i;
  passData->numStart=0;
  passData->numMiddle=0;
  passData->numEnd=0;
  passData->numWords=0;
  passData->numLayout=0;
  passData->numNumbers=0;
  for (i=0; i<Max_Input_Size; i++) {
    passData->startCount[i]=0;
    passData->middleCount[i]=0;
    passData->endCount[i]=0;
  }
  for (i=0; i<Max_Words; i++) {
    passData->wordCount[i]=0;
    passData->layoutCount[i]=0;
    passData->numberCount[i]=0;
  }
  return;
}

int parseInput(dataType * passData, char * nextString) {
  char tempString[Max_Error];
  char tempLayout[Max_Error];
  int count =0;
  int curCount=0;  //Used to mark the location of the character you
                   //are examining in the password

  int pos = 1;  //Used to tell where you are in the password
                //Added to store special chars/numbers that
                //occur in the begining, middle and end of passwords
  int curSpecial=0;
  int curWord=0;
  int curNumber=0;
  char tempChar;    //General temp variable, don't expect it to be initilized

  char letters[53];  //Used for comparing if a character is a letter
  char numbers[11];  //Used to compare if a character is a number
  char upperCase[27];  //Used to compare if a character is uppercase

  strcpy(upperCase,"ABCDEFGHIJKLMNOPQRSTUVWXYZ");
  strcpy(letters,"abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");
  strcpy(numbers,"1234567890");

  ///////////////////////////////
  //First find out what the password starts with.  Yes this was
  //added later instead of being rolled in with the secion below
  tempChar=nextString[0];
  if (strchr(letters,tempChar)!=NULL) {
    curWord=1;
    pos=0;
  }
  else if (strchr(numbers,tempChar)!=NULL) {
    curNumber=1;
    pos=0;
  }
  else { 
    curSpecial=1;
    pos=0;
  }
  tempString[curCount]=tempChar;
  if (strchr(letters,tempChar)!=NULL) {
    if (strchr(upperCase,tempChar)!=NULL) {
      tempLayout[count]='L';
    }
    else {
      tempLayout[count]='l';
    }
  }
  else if (strchr(numbers,tempChar)!=NULL) {
    tempLayout[count]='n';
  }
  else {
    tempLayout[count]='s';
  }
  count++;
  curCount++;
  tempChar=nextString[count]; 
     
  while (tempChar!='\0') {
    if (strchr(letters,tempChar)!=NULL) {
      if (strchr(upperCase,tempChar)!=NULL) {
        tempLayout[count]='L';
      }
      else {
        tempLayout[count]='l';
      }
      ////////////////////////////////////
      //Used to parse the special chars/numbers at the beg/middle/end
      //If the previous character was special, and now it's letters
      //stop recording the special characters
      if (curWord==0) {
        tempString[curCount]='\0';
        if (curSpecial==1) {
          insertString(passData, tempString, pos, 's'); 
        }
        else if (curNumber==0) {
          insertString(passData, tempString, pos, 'n');
        }
        curWord=1;
        curSpecial=0;
        curNumber=0;
        curCount=0;
        pos=1;
      }
      tempString[curCount]=tempChar;
      curCount++;
    }
    else if (strchr(numbers,tempChar)!=NULL) {
      tempLayout[count]='n';
      if (curNumber==0) {
        tempString[curCount]='\0';
        if (curSpecial==1) {
          insertString(passData, tempString, pos, 's');
        }
        else if (curWord==1) {
          insertString(passData, tempString, pos, 'w');
        }
        curWord=0;
        curSpecial=0;
        curNumber=1;
        curCount=0;
        pos=1;
      }
      tempString[curCount]=tempChar;
      curCount++;
    }
    else {
      tempLayout[count]='s';
      if (curSpecial==0) {
        tempString[curCount]='\0';
        if (curWord==1) {
          insertString(passData, tempString, pos, 'w');
        }
        else if (curNumber==1) {
          insertString(passData, tempString, pos, 'n');
        }
        curWord=0;
        curNumber=0;
        curSpecial=1;
        curCount=0;
        pos=1;
      }
      tempString[curCount]=tempChar;
      curCount++;
    }
    count++;
    tempChar=nextString[count];
  }
  tempLayout[count]='\0';
  tempString[curCount]='\0';
  insertString(passData,tempLayout,3,'l'); 

  if (curWord==1) {
    insertString(passData, tempString, 2, 'w');
  }
  else if (curNumber==1) {
    insertString(passData, tempString, 2, 'n');
  }
  else {
    insertString(passData, tempString, 2, 's');
  }

/*
  if (curWord==1) {
    insertString(passData, tempString, 1, 'w');
  }
  else if (curNumber==1) {
    insertString(passData, tempString, 1, 'n');
  }
  else {
    insertString(passData, tempString, 1, 's');
  }
*/

  return 0;
} 


/////////////////////////////////////////////////////////////////////
//
// Takes 4 inputtypes for wType
// w: letters(either upper or lowercase)
// n: numbers
// s: special characters
// l: layout type
//
// pos: the location of the data
// 0: Start of the password
// 1: middle of the password
// 2: end of the password
//
void insertString(dataType * passData, char * nextString, int pos, char wType) {
  int i;
  int done=0;
  char temp;
  if (wType=='w') {

    //////////////////
    //Lower case everything for saving the words
    //Case info is still saved in the layouts, this is for
    //generating word lists
    //
    for (i=0;i<Pass_Length;i++) {
      temp = nextString[i]; 
      if (temp=='\0') {
        i=Pass_Length;
      }
      else {
        nextString[i]=tolower(nextString[i]);
      }
    }

    for (i=0;i<passData->numWords;i++) {
      if (strcmp(nextString,passData->words[i])==0) {
        passData->wordCount[i]++;
        i=passData->numWords;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numWords;
      strcpy(passData->words[i],nextString);
      passData->numWords++;
      if (passData->numWords==(Max_Words-2)) {
        printf("Error, there are too many unique words to parse\n");
        printf("Increase the max word size and recompile\n");
        passData->numWords--;
      }
      passData->wordCount[i]=1;
    }
    
  }
  else if (wType=='n') {
    for (i=0;i<passData->numNumbers;i++) {
      if (strcmp(nextString,passData->numbers[i])==0) {
        passData->numberCount[i]++;
        i=passData->numNumbers;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numNumbers;
      strcpy(passData->numbers[i],nextString);
      passData->numNumbers++;
      if (passData->numNumbers==(Max_Words-2)) {
        printf("Error, there are too many unique numbers to parse\n");
        printf("Increase the max number size and recompile\n");
        passData->numNumbers--;
      }
      passData->numberCount[i]=1;
    }
  }
  else if (wType=='l') {
    for (i=0;i<passData->numLayout;i++) {
      if (strcmp(nextString,passData->layout[i])==0) {
        passData->layoutCount[i]++;
        i=passData->numLayout;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numLayout;
      strcpy(passData->layout[i],nextString);
      passData->numLayout++;
      if (passData->numLayout==(Max_Words-2)) {
        printf("Error, there are too many unique layouts to parse\n");
        printf("Increase the max layout size and recompile\n");
        passData->numLayout--;
      }

      passData->layoutCount[i]=1;
    }
  }
  if (((wType=='s')||(wType=='n'))&&(pos==0)) {
    for (i=0;i<passData->numStart;i++) {
      if (strcmp(nextString,passData->startSpecial[i])==0) {
        passData->startCount[i]++;
        i=passData->numStart;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numStart;
      strcpy(passData->startSpecial[i],nextString);
      passData->numStart++;
      if (passData->numStart==(Max_Input_Size-2)) {
        printf("Error, there are too many unique startChars to parse\n");
        printf("Increase the Max_Input_Size and recompile\n");
        passData->numStart--;
      }
      passData->startCount[i]=1;
    }
  }
  else if (((wType=='s')||(wType=='n'))&&(pos==1)) {
    for (i=0;i<passData->numMiddle;i++) {
      if (strcmp(nextString,passData->middleSpecial[i])==0) {
        passData->middleCount[i]++;
        i=passData->numMiddle;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numMiddle;
      strcpy(passData->middleSpecial[i],nextString);
      passData->numMiddle++;
      if (passData->numMiddle==(Max_Input_Size-2)) {
        printf("Error, there are too many unique middleChars to parse\n");
        printf("Increase the Max_Input_Size and recompile\n");
        passData->numMiddle--;
      }
      passData->middleCount[i]=1;
    }
  }
  else if (((wType=='s')||(wType=='n'))&&(pos==2)){
    for (i=0;i<passData->numEnd;i++) {
      if (strcmp(nextString,passData->endSpecial[i])==0) {
        passData->endCount[i]++;
        i=passData->numEnd;
        done=1;
      }
    }
    if (done!=1) {
      i = passData->numEnd;
      strcpy(passData->endSpecial[i],nextString);
      passData->numEnd++;
      if (passData->numEnd==(Max_Input_Size-2)) {
        printf("Error, there are too many unique endChars to parse\n");
        printf("Increase the Max_Input_Size and recompile\n");
        passData->numEnd--;
      }
      passData->endCount[i]=1;
    }

  }
  return;
}

void printOutput(FILE * startFile, FILE * middleFile, FILE * endFile, FILE* wordFile, FILE* layoutFile, FILE* numberFile,dataType * passData) {
  int i=0;
  int j=0;
  int curMax=0;
  char newSpecial[Max_Words][Pass_Length+1];
  int newCount[Max_Words];
  for (i=0; i<passData->numStart; i++) {
    for (j=0; j<passData->numStart; j++) {
      if (passData->startCount[j]>passData->startCount[curMax]) {
        curMax=j;
      }
    }
    if ((curMax<0)||(curMax>=Max_Words)) {
      printf("ERROR\n");
    }
    strcpy(newSpecial[i],passData->startSpecial[curMax]);
    newCount[i]=passData->startCount[curMax];
    passData->startCount[curMax]=-1;
    curMax=0;
  }


  for (i=0; i<passData->numStart; i++) {
    if (strlen(newSpecial[i])>7) {
      fprintf(startFile,"%s\t%i\n",newSpecial[i],newCount[i]);
    }
    else {
      fprintf(startFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
    }
  }

  curMax=0;
  for (i=0; i<passData->numMiddle; i++) {
    for (j=0; j<passData->numMiddle; j++) {
      if (passData->middleCount[j]>passData->middleCount[curMax]) {
        curMax=j;
      }
    }
    strcpy(newSpecial[i],passData->middleSpecial[curMax]);
    newCount[i]=passData->middleCount[curMax];
    passData->middleCount[curMax]=-1;
    curMax=0;
  }




  for (i=0; i<passData->numMiddle; i++) {
    if (strlen(newSpecial[i])>7) {
      fprintf(middleFile,"%s\t%i\n",newSpecial[i],newCount[i]);
    }
    else {
      fprintf(middleFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
    }
  }

  curMax=0;
  for (i=0; i<passData->numEnd; i++) {
    for (j=0; j<passData->numEnd; j++) {
      if (passData->endCount[j]>passData->endCount[curMax]) {
        curMax=j;
      }
    }
    strcpy(newSpecial[i],passData->endSpecial[curMax]);
    newCount[i]=passData->endCount[curMax];
    passData->endCount[curMax]=-1;
    curMax=0;
  }
  for (i=0; i<passData->numEnd; i++) {
    if (strlen(newSpecial[i])>7) {
      fprintf(endFile,"%s\t%i\n",newSpecial[i],newCount[i]);
    }
    else {
     fprintf(endFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
    }
  }

  curMax=0;
  for (i=0; i<passData->numWords; i++) {
    for (j=0; j<passData->numWords; j++) {
      if (passData->wordCount[j]>passData->wordCount[curMax]) {
        curMax=j;
      }
    }
    strcpy(newSpecial[i],passData->words[curMax]);
    newCount[i]=passData->wordCount[curMax];
    passData->wordCount[curMax]=-1;
    curMax=0;
  }
  for (i=0; i<passData->numWords; i++) {
    if (strlen(newSpecial[i])>7) {
      if (printOnlyWords==1) {
        fprintf(wordFile,"%s\n",newSpecial[i]);
      }
      else {
        fprintf(wordFile,"%s\t%i\n",newSpecial[i],newCount[i]);
      }
    }
    else {
      if (printOnlyWords==1) {
        fprintf(wordFile,"%s\n",newSpecial[i],newCount[i]);
      }
      else {
        fprintf(wordFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
      }
    }
  }


  curMax=0;
  for (i=0; i<passData->numNumbers; i++) {
    for (j=0; j<passData->numNumbers; j++) {
      if (passData->numberCount[j]>passData->numberCount[curMax]) {
        curMax=j;
      }
    }
    strcpy(newSpecial[i],passData->numbers[curMax]);
    newCount[i]=passData->numberCount[curMax];
    passData->numberCount[curMax]=-1;
    curMax=0;
  }
  for (i=0; i<passData->numNumbers; i++) {
    if (strlen(newSpecial[i])>7) {
      fprintf(numberFile,"%s\t%i\n",newSpecial[i],newCount[i]);
    }
    else {
      fprintf(numberFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
    }
  }



  curMax=0;
  for (i=0; i<passData->numLayout; i++) {
    for (j=0; j<passData->numLayout; j++) {
      if (passData->layoutCount[j]>passData->layoutCount[curMax]) {
        curMax=j;
      }
    }
    strcpy(newSpecial[i],passData->layout[curMax]);
    newCount[i]=passData->layoutCount[curMax];
    passData->layoutCount[curMax]=-1;
    curMax=0;
  }


  for (i=0; i<passData->numLayout; i++) {
    if (strlen(newSpecial[i])>7) {
      fprintf(layoutFile,"%s\t%i\n",newSpecial[i],newCount[i]);
    }
    else {
      fprintf(layoutFile,"%s\t\t%i\n",newSpecial[i],newCount[i]);
    }
  }




  return;
}
